/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include "backend.h"
#include "../owsip.h"
#include "../call/call.h"
#include "../dial/detector.h"

enum state
{
	ONHOOK = 0,	// on hook
	RING,		// on hook ; ringing
	OFFHOOK,	// off hook
	DIAL,		// off hook ; started dialing
	CONNECT,	// off hook ; call is connecting
	CALL,		// off hook ; in a call
	END,		// off hook ; call ended
	ERROR,		// error occured
};

struct ltq_tapi_port
{
	enum state state;	// the state of the port
	struct call *call;	// the call that we are handling, if any
	struct detector detector;	// we feed the detector digits and get numbers
	pj_timer_entry ring_timer;	// the timer used to generate the ring intervals
};

extern void (*tapi_digit_callback)(unsigned int port, unsigned char digit);
extern void (*tapi_hook_callback)(unsigned int port, unsigned char event);

extern pj_status_t tapi_hook_status(pj_uint32_t port, pj_uint32_t *status);
extern pj_status_t tapi_dial_tone(pj_uint32_t port);
extern pj_status_t tapi_no_tone(pj_uint32_t port);
extern pj_status_t tapi_ring(pj_uint32_t port, pj_uint32_t state, char *caller_number);

#define MAX_PORT 2

static struct ltq_tapi_port ports[MAX_PORT];

static void
ltq_tapi_pickup(struct call *call, int port)
{
	OWSIP_TRACER("enter");

	int i;
	for(i = 0; i < MAX_PORT; i++)
	{
		if((ports[i].state == RING) && (ports[i].call == call))
		{
			tapi_ring(i, 0, NULL);
			if(i == port)
			{
				ports[i].state = CONNECT;
				owsip_pickup(ports[i].call->call_id);
			} else {
				ports[i].state = ONHOOK;
				ports[i].call = NULL;
			}
		}
	}
}

static void
ltq_tapi_detector_event(struct detector *d, enum detector_event e, char data)
{
	OWSIP_TRACER("enter");

	int port = (int)d->priv;

	switch(e) {
		case DETECTOR_HOOK:
			printf("%s hook port %d\n", (data)?("off"):("on"), port);
			if(data) {
				if((ports[port].state == RING) && ports[port].call) {
					ltq_tapi_pickup(ports[port].call, port);
				} else {
					ports[port].state = OFFHOOK;
					tapi_dial_tone(port);
				}
			} else {
				if(((ports[port].state == CALL) || (ports[port].state == CONNECT)) && ports[port].call) {
					owsip_hangup(ports[port].call->call_id);
					ports[port].call = 0;
				}
				tapi_no_tone(port);
				ports[port].state = ONHOOK;
			}
			break;

		case DETECTOR_DIGIT:
			break;
	}
}

static void
ltq_tapi_detector_number(struct detector *d, char *nr)
{ 
	OWSIP_TRACER("enter");

	pj_status_t status;
	int port = (int)d->priv;
	status = owsip_call_realm(nr);
	if (status == PJ_SUCCESS)
		ports[port].state = CONNECT;
	else
		ports[port].state = ERROR;
	printf("port %d dialed %s\n", port, nr);
}

void ltq_tapi_digit(unsigned int port, unsigned char digit)
{
	OWSIP_TRACER("enter");

	ports[port].state = DIAL;
	tapi_no_tone(port);
	detector_feed(&ports[port].detector, DETECTOR_DIGIT, digit);
}

void ltq_tapi_hook(unsigned int port, unsigned char event)
{
	OWSIP_TRACER("enter");

	detector_feed(&ports[port].detector, DETECTOR_HOOK, event);
}

static void
ltq_tapi_init(struct owsip *owsip)
{
	OWSIP_TRACER("enter");

	unsigned int count = 16;
	pjmedia_aud_dev_info info[16];
	pj_uint32_t hook_status;
	int i;

	tapi_digit_callback = ltq_tapi_digit;
	tapi_hook_callback = ltq_tapi_hook;

	pjsua_enum_aud_devs(info, &count);
	printf("audio devices\n");
	for(i = 0; i < count; i++)
		printf("%d -> %s\n", i, info[i].name);
	pjsua_set_snd_dev(0, 0);

	for(i = 0; i < MAX_PORT; i++) {
		ports[i].detector.event = ltq_tapi_detector_event;
		ports[i].detector.number = ltq_tapi_detector_number;
		ports[i].detector.priv = (void*)i;
		detector_init(&ports[i].detector);
		tapi_hook_status(i, &hook_status);
		if (hook_status) {
			tapi_dial_tone(i);
			ports[i].state = OFFHOOK;
		}
	}
}

static void
ltq_tapi_connect(struct call *c)
{
	OWSIP_TRACER("enter");

	pjsua_call_info ci;
	pjsua_call_get_info(c->call_id, &ci);
	if (ci.media_status == PJSUA_CALL_MEDIA_ACTIVE) {
		pjsua_conf_connect(ci.conf_slot, 0);
		pjsua_conf_connect(0, ci.conf_slot);
	}
}

static void
ltq_tapi_disconnect(struct call *c)
{
	OWSIP_TRACER("enter");

	pjsua_call_info ci;
	pjsua_call_get_info(c->call_id, &ci);
	if (ci.media_status == PJSUA_CALL_MEDIA_ACTIVE) {
		pjsua_conf_disconnect(ci.conf_slot, 0);
	}
}

static void
ltq_tapi_media_config(pjsua_media_config *media_cfg)
{
	OWSIP_TRACER("enter");

	media_cfg->clock_rate = 8000;
	media_cfg->snd_auto_close_time = -1;
}

static void
ltq_tapi_start(struct call *c)
{
	OWSIP_TRACER("enter");

	int i;
	for(i = 0; i < MAX_PORT; i++) {
		if (ports[i].state == ONHOOK) {
			ports[i].call = c;
			pjsua_call_info ci;
			char caller_number[15 + 1] = "\0";
			pjsua_call_get_info(c->call_id, &ci);

			sscanf(ci.remote_info.ptr, "\"%15[^\"]s", caller_number);
			ports[i].state = RING;

			if (strncmp("Anonymous", caller_number, strlen("Anonymous")) != 0)
				tapi_ring(i, 1, caller_number);
			else
				tapi_ring(i, 1, NULL);
			break;
		}
		if (ports[i].state == CONNECT) {
			ports[i].call = c;
			ports[i].state = CALL;
			break;
		}
	}
	printf("setting call_id %d on port %d\n", c->call_id, i);
}

static void
ltq_tapi_stop(struct call *c)
{
	OWSIP_TRACER("enter");

	int i;
	for(i = 0; i < MAX_PORT; i++) {
		if(ports[i].call == c) {
			printf("stopping call %d on port %d\n", c->call_id, i);
			if (ports[i].state == RING) {
				ports[i].state = ONHOOK;
				tapi_ring(i, 0, NULL);
				break;
			}
			ports[i].state = END;
			break;
		}
	}
}

static struct backend ltq_tapi_backend = {
	.ports = MAX_PORT,
	.name = "ltq_tapi",
	.init = ltq_tapi_init,
	.connect = ltq_tapi_connect,
	.disconnect = ltq_tapi_disconnect,
	.media_config = ltq_tapi_media_config,
	.start = ltq_tapi_start,
	.stop = ltq_tapi_stop,
};

BACKEND(ltq_tapi_backend)
