/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include "../config.h"
#include "../owsip.h"
#include "call.h"

#define TOUT_CALL	30

static LIST_HEAD(call_handlers);

static struct call calls[PJSUA_MAX_CALLS];

void call_handler_register(struct call_handler *c)
{
	OWSIP_TRACER("enter");

	if(c->name)
		list_add(&c->list, &call_handlers);
	else
		printf("call_handler %s is not configured properly\n", c->name);
}

static struct call_handler*
call_handler_find(const char *name)
{
	OWSIP_TRACER("enter");

	struct list_head *p;
	if(!name)
		return 0;
	list_for_each(p, &call_handlers) {
		struct call_handler *c = container_of(p, struct call_handler, list);
		if(!strcmp(c->name, name))
			return c;
	}
	return 0;
}

static struct rule*
call_match_rule(struct owsip *owsip, pjsua_call_id call_id)
{
	OWSIP_TRACER("enter");

	/* magic call match based on rules goes here */
	return &owsip->policy;
}

static void
call_stop_cb(pj_timer_heap_t *timer_heap, struct pj_timer_entry *entry)
{
	OWSIP_TRACER("enter");

	struct call *c = (struct call*) entry->user_data;
	printf("call stop event, hanging up\n");
	owsip_hangup(c->call_id);
}

static void
call_start_cb(pj_timer_heap_t *timer_heap, struct pj_timer_entry *entry)
{
	OWSIP_TRACER("enter");

	struct call *c = (struct call*) entry->user_data;
	printf("call not answered on time, using call handler \n");
	if(c->rule->stop) {
		pj_time_val d;
		d.sec = c->rule->stop;
		d.msec = 0;
		calls[c->call_id].timer.cb = call_stop_cb;
		pjsua_schedule_timer(&calls[c->call_id].timer, &d);
	}
	if(!c->handler)
		return;
	if(c->handler->start)
		c->handler->start(c);
}

static void
call_start(struct owsip *owsip, pjsua_call_id call_id)
{
	OWSIP_TRACER("enter");

	if(calls[call_id].backend->start)
		calls[call_id].backend->start(&calls[call_id]);
	if(calls[call_id].rule->start) {
		pj_time_val d;
		d.sec = calls[call_id].rule->start;
		d.msec = 0;
		calls[call_id].timer.cb = call_start_cb;
		calls[call_id].timer.user_data = (void*) &calls[call_id];
		pjsua_schedule_timer(&calls[call_id].timer, &d);
	}
}

static void
call_connect(struct owsip *owsip, pjsua_call_id call_id)
{
	OWSIP_TRACER("enter");

	pjsua_cancel_timer(&calls[call_id].timer);
	if(calls[call_id].backend->connect)
		calls[call_id].backend->connect(&calls[call_id]);
}

static void
call_stop(struct owsip *owsip, pjsua_call_id call_id)
{
	OWSIP_TRACER("enter");

	pjsua_cancel_timer(&calls[call_id].timer);
	if(calls[call_id].backend->disconnect)
		calls[call_id].backend->disconnect(&calls[call_id]);
	if(calls[call_id].backend->stop)
		calls[call_id].backend->stop(&calls[call_id]);
}

void
call_state(struct owsip *owsip, pjsua_call_id call_id, int state)
{
	OWSIP_TRACER("enter");

	memset(&calls[call_id], 0, sizeof(struct call));
	calls[call_id].call_id = call_id;
	calls[call_id].backend = owsip->backend;
	calls[call_id].rule = call_match_rule(owsip, call_id);
	calls[call_id].handler = call_handler_find(calls[call_id].rule->handler);
	calls[call_id].transport = pjsua_call_get_media_transport(call_id);

	switch(state) {
		case CALL_START:
			call_start(owsip, call_id);
			break;
		case CALL_STOP:
			call_stop(owsip, call_id);
			break;
		case CALL_CONNECT:
			call_connect(owsip, call_id);
			break;
		default:
			break;
	}
}
