/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <libubox/list.h>

#include "config.h"
#include "dial/dial.h"
#include "call/call.h"

static struct owsip *owsip_ctx = 0;

static void
on_incoming_call(pjsua_acc_id acc_id, pjsua_call_id call_id, pjsip_rx_data *rdata)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	printf("Incoming call from %.*s !!!\n", (int)ci.remote_info.slen, ci.remote_info.ptr);
	pjsua_call_answer(call_id, 180, NULL, NULL);
	call_state(owsip_ctx, call_id, CALL_START);
}

static void
on_call_state(pjsua_call_id call_id, pjsip_event *e)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	printf("[%s] call:%d %.*s\n", __FILE__, call_id, (int)ci.state_text.slen, ci.state_text.ptr);
	switch(ci.state) {
	case PJSIP_INV_STATE_NULL:
		break;
	case PJSIP_INV_STATE_CALLING:
		call_state(owsip_ctx, call_id, CALL_START);
		break;
	case PJSIP_INV_STATE_INCOMING:
		break;
	case PJSIP_INV_STATE_EARLY:
		break;
	case PJSIP_INV_STATE_CONNECTING:
		break;
	case PJSIP_INV_STATE_CONFIRMED:
		break;
	case PJSIP_INV_STATE_DISCONNECTED:
		call_state(owsip_ctx, call_id, CALL_STOP);
		break;
	}
}

static void
on_call_media_state(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	call_state(owsip_ctx, call_id, CALL_CONNECT);
}

static void
on_nat_detect(const pj_stun_nat_detect_result *res)
{
	if(res->status != PJ_SUCCESS)
		pjsua_perror("owsip", "NAT detection failed", res->status);
	else
		printf("NAT detected as %s", res->nat_type_name);
}

static pj_status_t
owsip_init(struct owsip *owsip)
{
	pj_status_t status;
	pjsua_media_config default_media_cfg;

	owsip_ctx->backend = backend_get(owsip->conf.backend);
	if(!owsip_ctx->backend) {
		printf("failed to load backend %s\n", owsip->conf.backend);
		exit(1);
	} else {
		printf("using backend %s\n", owsip_ctx->backend->name);
	}

	status = pjsua_create();
	if (status != PJ_SUCCESS) {
		printf("pjsua_create failed\n");
		return status;
	}

	pjsua_config_default(&owsip->cfg);

	owsip->cfg.cb.on_incoming_call = &on_incoming_call;
	owsip->cfg.cb.on_call_media_state = &on_call_media_state;
	owsip->cfg.cb.on_call_state = &on_call_state;
	owsip->cfg.cb.on_nat_detect = &on_nat_detect;

	pjsua_logging_config_default(&owsip->log_cfg);
	owsip->log_cfg.console_level = owsip->conf.log_level;

	pjsua_media_config_default(&default_media_cfg);
	if(owsip_ctx->backend->media_config)
		owsip_ctx->backend->media_config(&default_media_cfg);

	if (!pjsua_verify_sip_url(owsip->conf.outbound)) {
		owsip->cfg.outbound_proxy[owsip->cfg.outbound_proxy_cnt++] = pj_str(owsip->conf.outbound);
	}

	status = pjsua_init(&owsip->cfg, (owsip->conf.log_level)?(&owsip->log_cfg):(NULL), &default_media_cfg);
	if (status != PJ_SUCCESS) {
		printf("pjsua_init failed\n");
		return status;
	}

	// TODO: handle case when IP changes (this should be function by it's own)
	char *interface_ip = NULL;
	if (owsip->conf.interface) {
		struct ifreq ifr;
		int fd, status;

get_interface_ip:
		fd = socket(AF_INET, SOCK_DGRAM, 0);
		ifr.ifr_addr.sa_family = AF_INET;
		strncpy(ifr.ifr_name, owsip->conf.interface, IFNAMSIZ - 1);
		status = ioctl(fd, SIOCGIFADDR, &ifr);
		close(fd);

		if (!status) {
			interface_ip = strdup(inet_ntoa(((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr));
		} else {
			// wait until we got right interface IP
			printf("interface without IP retrying\n");
			sleep(5);
			goto get_interface_ip;
		}
	}

	pjsua_transport_config udp_cfg;
	pjsua_transport_config_default(&udp_cfg);

	if (owsip->conf.local_port) {
		udp_cfg.port = owsip->conf.local_port;
	} else {
		udp_cfg.port = 5060;

	}

	if (interface_ip) {
		udp_cfg.public_addr = pj_str(interface_ip);
		udp_cfg.bound_addr = pj_str(interface_ip);
	}

	status = pjsua_transport_create(PJSIP_TRANSPORT_UDP, &udp_cfg, NULL);
	if (status != PJ_SUCCESS) {
		printf("pjsua_transport_create() failed\n");
		return status;
	}

	pjsua_transport_config rtp_cfg;
	pjsua_transport_config_default(&rtp_cfg);

	if (owsip->conf.rtp_port) {
		rtp_cfg.port = owsip->conf.rtp_port;
	} else {
		rtp_cfg.port = 4000;
	}

	if (interface_ip) {
		rtp_cfg.public_addr = pj_str(interface_ip);
		rtp_cfg.bound_addr = pj_str(interface_ip);
	}

	status = pjsua_media_transports_create(&rtp_cfg);
	if (status != PJ_SUCCESS) {
		printf("pjsua_media_transports_create() failed\n");
		return status;
	}

	status = pjsua_call_subsys_start();
	if (status != PJ_SUCCESS)
		return status;

	status = pjsua_media_subsys_start();
	if (status != PJ_SUCCESS)
		return status;

	if(owsip_ctx->backend->init)
		owsip_ctx->backend->init(owsip_ctx);

	dial_state(STATE_INIT);

	status = pjsua_pres_start();
	return status;
}

static pj_status_t
owsip_account(struct owsip *owsip)
{
	char id[512];
	char reg_uri[256];
	struct list_head *p;
	pj_status_t status;

	list_for_each(p, &owsip->accounts) {
		struct account *account = list_entry(p, struct account, list);

		pjsua_acc_config_default(&account->acc_config);

		snprintf(id, 512, "sip:%s@%s", account->username, account->realm);
		snprintf(reg_uri, 256, "sip:%s", account->realm);

		account->acc_config.id = pj_str(id);
		account->acc_config.reg_uri = pj_str(reg_uri);
		account->acc_config.cred_count = 1;
		account->acc_config.cred_info[0].realm = pj_str(account->realm);
		account->acc_config.cred_info[0].scheme = pj_str("digest");
		account->acc_config.cred_info[0].username = pj_str(account->username);
		account->acc_config.cred_info[0].data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
		account->acc_config.cred_info[0].data = pj_str(account->password);

		status = pjsua_acc_add(&account->acc_config, PJ_TRUE, &account->acc_id);
		if (status != PJ_SUCCESS)
			return status;

		/* until rules are not finished use last account as default one */
		status = pjsua_acc_set_default(account->acc_id);
		if (status != PJ_SUCCESS)
			return status;
	}

	return PJ_SUCCESS;
}

pj_status_t
owsip_call(char *curi)
{
	/* until rules are not finished use last account as default one */
	pj_str_t uri;
	pj_status_t status = pjsua_verify_url(curi);
	if (status != PJ_SUCCESS) {
		printf("pjsip says that %s is not a valid sip uri. not calling it\n", curi);
		return status;
	}
	uri = pj_str(curi);
	return pjsua_call_make_call(pjsua_acc_get_default(), &uri, 0, NULL, NULL, NULL);
}

pj_status_t
owsip_call_realm(char *curi)
{
	/* until rules are not finished use last account as default one */
	char buf[512];
	struct list_head *p;

	list_for_each(p, &owsip_ctx->accounts) {
		struct account *account = list_entry(p, struct account, list);
		snprintf(buf, 512, "sip:%s@%s", curi, account->realm);
	}
	return owsip_call(buf);
}

pj_status_t
owsip_pickup(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	printf("answering call from %.*s!!", (int)ci.remote_info.slen, ci.remote_info.ptr);
	return pjsua_call_answer(call_id, 200, NULL, NULL);
}

pj_status_t
owsip_hangup(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	printf("hanging up call with %.*s!!", (int)ci.remote_info.slen, ci.remote_info.ptr);
	return pjsua_call_hangup(call_id, 410, NULL, NULL);
}

static void
mainloop(void)
{
	int loop = 1;

	while(loop) {
		char option[256];
		gets(option);
		switch(*option) {
			case 'a':
				{
					unsigned int cnt = 1;
					pjsua_call_id id;
					pjsua_enum_calls(&id, &cnt);
					if(cnt) {
						printf("answering call %d\n", id);
						owsip_pickup(id);
					}
				}
				break;

			case 'c':
				dial_call_nr(&option[1]);
				break;

			case 'h':
				pjsua_call_hangup_all();
				break;

			case 'q':
				loop = 0;
				break;
		}
	}
}

static void
sigterm(int signum)
{
	printf("received sigterm\n");
	pjsua_destroy();
	exit(-1);
}

int
main(int argc, char *argv[])
{
	signal(SIGTERM, sigterm);
	signal(SIGINT, sigterm);

	pj_status_t status;

	owsip_ctx = config_load();
	status = owsip_init(owsip_ctx);
	status = owsip_account(owsip_ctx);

	if(owsip_ctx->backend->mainloop)
		owsip_ctx->backend->mainloop(owsip_ctx);
	else
		mainloop();

	pjsua_destroy();
	return 0;
}
